@@ -10,10 +10,13 @@ from copy import copy
 
 
 class Pattern(Serialize):
+    raw = None
+    type = None
 
-    def __init__(self, value, flags=()):
+    def __init__(self, value, flags=(), raw=None):
         self.value = value
         self.flags = frozenset(flags)
+        self.raw = raw
 
     def __repr__(self):
         return repr(self.to_regexp())
@@ -28,6 +31,12 @@ class Pattern(Serialize):
     def to_regexp(self):
         raise NotImplementedError()
 
+    def min_width(self):
+        raise NotImplementedError()
+
+    def max_width(self):
+        raise NotImplementedError()
+
     if Py36:
         # Python 3.6 changed syntax for flags in regular expression
         def _get_flags(self, value):
@@ -42,6 +51,7 @@ class Pattern(Serialize):
             return value
 
 
+
 class PatternStr(Pattern):
     __serialize_fields__ = 'value', 'flags'
 
@@ -92,6 +102,12 @@ class TerminalDef(Serialize):
     def __repr__(self):
         return '%s(%r, %r)' % (type(self).__name__, self.name, self.pattern)
 
+    def user_repr(self):
+        if self.name.startswith('__'): # We represent a generated terminal
+            return self.pattern.raw or self.name
+        else:
+            return self.name
+
 
 class Token(Str):
     """A string with meta-information, that is produced by the lexer.
@@ -283,7 +299,7 @@ class Lexer(object):
 class TraditionalLexer(Lexer):
 
     def __init__(self, conf):
-        terminals = list(conf.tokens)
+        terminals = list(conf.terminals)
         assert all(isinstance(t, TerminalDef) for t in terminals), terminals
 
         self.re = conf.re_module
@@ -299,7 +315,8 @@ class TraditionalLexer(Lexer):
                 if t.pattern.min_width == 0:
                     raise LexError("Lexer does not allow zero-width terminals. (%s: %s)" % (t.name, t.pattern))
 
-            assert set(conf.ignore) <= {t.name for t in terminals}
+            if not (set(conf.ignore) <= {t.name for t in terminals}):
+                raise LexError("Ignore terminals are not defined: %s" % (set(conf.ignore) - {t.name for t in terminals}))
 
         # Init
         self.newline_types = frozenset(t.name for t in terminals if _regexp_has_newline(t.pattern.to_regexp()))
@@ -310,6 +327,7 @@ class TraditionalLexer(Lexer):
         self.user_callbacks = conf.callbacks
         self.g_regex_flags = conf.g_regex_flags
         self.use_bytes = conf.use_bytes
+        self.terminals_by_name = conf.terminals_by_name
 
         self._mres = None
 
@@ -338,12 +356,12 @@ class TraditionalLexer(Lexer):
             if m:
                 return m.group(0), type_from_index[m.lastindex]
 
-    def lex(self, state, _parser_state):
+    def lex(self, state, parser_state):
         with suppress(EOFError):
             while True:
-                yield self.next_token(state)
+                yield self.next_token(state, parser_state)
 
-    def next_token(self, lex_state):
+    def next_token(self, lex_state, parser_state=None):
         line_ctr = lex_state.line_ctr
         while line_ctr.char_pos < len(lex_state.text):
             res = self.match(lex_state.text, line_ctr.char_pos)
@@ -352,7 +370,8 @@ class TraditionalLexer(Lexer):
                 if not allowed:
                     allowed = {"<END-OF-FILE>"}
                 raise UnexpectedCharacters(lex_state.text, line_ctr.char_pos, line_ctr.line, line_ctr.column,
-                                           allowed=allowed, token_history=lex_state.last_token and [lex_state.last_token])
+                                           allowed=allowed, token_history=lex_state.last_token and [lex_state.last_token],
+                                           state=parser_state, terminals_by_name=self.terminals_by_name)
 
             value, type_ = res
 
@@ -365,7 +384,7 @@ class TraditionalLexer(Lexer):
                 if t.type in self.callback:
                     t = self.callback[t.type](t)
                     if not isinstance(t, Token):
-                        raise ValueError("Callbacks must return a token (returned %r)" % t)
+                        raise LexError("Callbacks must return a token (returned %r)" % t)
                 lex_state.last_token = t
                 return t
             else:
@@ -393,14 +412,11 @@ class LexerState:
 class ContextualLexer(Lexer):
 
     def __init__(self, conf, states, always_accept=()):
-        terminals = list(conf.tokens)
-        tokens_by_name = {}
-        for t in terminals:
-            assert t.name not in tokens_by_name, t
-            tokens_by_name[t.name] = t
+        terminals = list(conf.terminals)
+        terminals_by_name = conf.terminals_by_name
 
         trad_conf = copy(conf)
-        trad_conf.tokens = terminals
+        trad_conf.terminals = terminals
 
         lexer_by_tokens = {}
         self.lexers = {}
@@ -410,15 +426,14 @@ class ContextualLexer(Lexer):
                 lexer = lexer_by_tokens[key]
             except KeyError:
                 accepts = set(accepts) | set(conf.ignore) | set(always_accept)
-                state_tokens = [tokens_by_name[n] for n in accepts if n and n in tokens_by_name]
                 lexer_conf = copy(trad_conf)
-                lexer_conf.tokens = state_tokens
+                lexer_conf.terminals = [terminals_by_name[n] for n in accepts if n in terminals_by_name]
                 lexer = TraditionalLexer(lexer_conf)
                 lexer_by_tokens[key] = lexer
 
             self.lexers[state] = lexer
 
-        assert trad_conf.tokens is terminals
+        assert trad_conf.terminals is terminals
         self.root_lexer = TraditionalLexer(trad_conf)
 
     def make_lexer_state(self, text):
@@ -428,15 +443,18 @@ class ContextualLexer(Lexer):
         try:
             while True:
                 lexer = self.lexers[parser_state.position]
-                yield lexer.next_token(lexer_state)
+                yield lexer.next_token(lexer_state, parser_state)
         except EOFError:
             pass
         except UnexpectedCharacters as e:
             # In the contextual lexer, UnexpectedCharacters can mean that the terminal is defined, but not in the current context.
             # This tests the input against the global context, to provide a nicer error.
-            token = self.root_lexer.next_token(lexer_state)
-            raise UnexpectedToken(token, e.allowed, state=parser_state.position)
-
+            try:
+                last_token = lexer_state.last_token  # Save last_token. Calling root_lexer.next_token will change this to the wrong token
+                token = self.root_lexer.next_token(lexer_state, parser_state)
+                raise UnexpectedToken(token, e.allowed, state=parser_state, token_history=[last_token], terminals_by_name=self.root_lexer.terminals_by_name)
+            except UnexpectedCharacters:
+                raise e  # Raise the original UnexpectedCharacters. The root lexer raises it with the wrong expected set.
 
 class LexerThread:
     """A thread that ties a lexer instance and a lexer state, to be used by the parser"""
